using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Pentium4emulator
{
    public partial class Form1 : Form
    {
        //My vars.
        //A dictionary that keeps track of labels and their line numbers.
        public Dictionary<string, int> labels=new Dictionary<string, int>();
        public p4engine p4;
        public myRam mr;
        MyParser parser;
        public RichTextBox programbox;
        public ToolStripStatusLabel statusbar;
        private string filename = "";
        private bool filesavedonce = false;
        const int stackinit = 44024;

        private Point currentLocation = new Point(0,0);

        public Form1() {
            InitializeComponent();
            statusBarItem1.Image = imageList1.Images[0];
            mr = new myRam();
            p4 = new p4engine(mr,this);
            programbox = richTextBox1;
            parser = new MyParser(this);
            statusbar = this.statusBarItem1;
            tabClicked(null, null);
			this.myTableViewer1.minimiseRequested += new MyTableViewer.myEventHandler(minimiseTabRequest);
			this.quickConsoleControl1.minimiseRequested += new MyTableViewer.myEventHandler(minimiseTabRequest);
			this.myTableViewer1.cellDoubleClicked += new MyTableViewer.showLocation(moveCursorToErrorLine);
			this.parser.compileError += new MyParser.errorOccured(parser_compileError);
            //richTextBox1.ZoomFactor = 1.5F;
        }

		void parser_compileError(int errormessage, int lineNumber, string param, ErrorTypes type) {
			string msg="";
			try {
				msg = (string)ErrorMessages.ResourceManager.GetObject("E" + errormessage) + param;
			}
			catch {
				msg = (string)ErrorMessages.ResourceManager.GetObject("E0");//Unknown Error
			}

			switch(type) {
				case ErrorTypes.Error:
					this.myTableViewer1.addError(lineNumber, msg);
					break;
				case ErrorTypes.Warning:
					this.myTableViewer1.addWarning(lineNumber, msg);
					break;
				case ErrorTypes.Message:
					this.myTableViewer1.addMessage(lineNumber, msg);
					break;
			}
			showOrHideBottomPane(0, true);
		}


        /// <summary>
        /// When the user clicks the "Registers" button, toggle the registers
        /// panel view.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button1_Click(object sender, EventArgs e) {
            regPanel.Visible = (!regPanel.Visible);
            if(button1.ImageIndex == 4) button1.ImageIndex = 5;
            else button1.ImageIndex = 4;
        }


        //Handler for the "Compile" Event
        private void btnCompile_Click(object sender, EventArgs e) {

			//Get rid of any pink background
			int selstart = richTextBox1.SelectionStart;
			richTextBox1.SelectAll();
			richTextBox1.SelectionBackColor = richTextBox1.BackColor;
			richTextBox1.DeselectAll();
			richTextBox1.SelectionStart = selstart;

			//Initialise / Empty the resources.
            labels.Clear();
            mr.Clear();
			myTableViewer1.clear();

            statusBarItem1.Image = imageList1.Images[1];
			if(!parser.getGlobals()) {
				//Globals not got correctly
				statusBarItem1.Image = imageList1.Images[3];
				statusBarItem1.Text = "Compile Error: Global variables could not be resolved";
				return;
			}
			//Find the labels and their corresponding line numbers
            parser.findLabels();

            //find  a way to accept both main and _main.
            if(!(labels.ContainsKey("_main") || labels.ContainsKey("_main"))) {
                //handleerror("Main function not found. Declare a label with the name '_main' or '_main' to declare a main function");
				parser_compileError(6, 0, "", ErrorTypes.Error);
				statusBarItem1.Image = imageList1.Images[3];
				statusBarItem1.Text = "Compile Error: Entry point not defined.";
				return;
            }
            p4.setReg(p4engine.regtype.epc, labels["_main"]);
            p4.setReg(p4engine.regtype.esp, stackinit); // change this should stkOffset change.
            //MessageBox.Show("set epc to: " + labels["_main"]);

            refreshRegisterList();
            refreshStackList();
            panel1.Refresh();//repaint the arrow
            statusBarItem1.Image = imageList1.Images[2];
            statusBarItem1.Text = "Compiled Successfully";

        }

     

        private void compileToolStripMenuItem_Click(object sender, EventArgs e) {
            btnCompile_Click(null, null);
        }


        internal void handleerror(string p) {
            //MessageBox.Show(this, p, "Compiler error");
            this.myTableViewer1.addError(0,p);
			showOrHideBottomPane(0, true);
            statusBarItem1.Image = imageList1.Images[3];
        }


        //handler for the "Step One Line" Command
        private void button2_Click(object sender, EventArgs e) {
            string rawline, goodcommand;
            while(true) {
                int line = p4.getReg(p4engine.regtype.epc);
                if(line >= richTextBox1.Lines.Length) {
                    MessageBox.Show("End of Program Reached.");
                    return;
                }
                if(labels.ContainsValue(line+1) && labels["_main"]!= (line+1)) {
                    p4.setReg(p4engine.regtype.epc, p4.getReg(p4engine.regtype.epc) + 1);
                    continue;
                }
                rawline = richTextBox1.Lines[line];
                goodcommand = getUsefulCommand(rawline);
                if(goodcommand.Length < 1) {
                    p4.setReg(p4engine.regtype.epc, p4.getReg(p4engine.regtype.epc) + 1);
                    continue;
                }
                else {
                    statusBarItem1.Text = "Line Number: " + (p4.getReg(p4engine.regtype.epc) + 1);
                    break;
                }
            }
             
            //p4.executeOpcode(goodcommand);
			parser.parseAndExecuteLine(goodcommand,p4.getReg(p4engine.regtype.epc)+1);
            /*int nextindex = richTextBox1.GetFirstCharIndexFromLine(p4.getReg(p4engine.regtype.epc));
            if(nextindex!=-1)
                richTextBox1.Select(nextindex, 0);
            richTextBox1.ScrollToCaret();*/
            panel1.Refresh();
            refreshRegisterList();
            refreshStackList();
        }

        private void refreshRegisterList() {
            if(mnuHexRegisters.Checked) {
                regeax.Text = p4.getReg(p4engine.regtype.eax).ToString("x");
                regebp.Text = p4.getReg(p4engine.regtype.ebp).ToString("x");
                regebx.Text = p4.getReg(p4engine.regtype.ebx).ToString("x");
                regecx.Text = p4.getReg(p4engine.regtype.ecx).ToString("x");
                regedi.Text = p4.getReg(p4engine.regtype.edi).ToString("x");
                regedx.Text = p4.getReg(p4engine.regtype.edx).ToString("x");
                regesi.Text = p4.getReg(p4engine.regtype.esi).ToString("x");
                regesp.Text = p4.getReg(p4engine.regtype.esp).ToString("x");
            }
            else {
                regeax.Text = p4.getReg(p4engine.regtype.eax) + "";
                regebp.Text = p4.getReg(p4engine.regtype.ebp) + "";
                regebx.Text = p4.getReg(p4engine.regtype.ebx) + "";
                regecx.Text = p4.getReg(p4engine.regtype.ecx) + "";
                regedi.Text = p4.getReg(p4engine.regtype.edi) + "";
                regedx.Text = p4.getReg(p4engine.regtype.edx) + "";
                regesi.Text = p4.getReg(p4engine.regtype.esi) + "";
                regesp.Text = p4.getReg(p4engine.regtype.esp) + "";
            }
        }

		//Refreshes the value of the stack elements...
		//Enormous scope for optimisation here, current method gets rid of all labels
		//and rebuilds it from scratch.
        private void refreshStackList() {
            if(!stackPanelHolder.Visible) return;
			//If Stack Pointer out of range
			//TODO: Sync this with RAM limits
            if(p4.getReg(p4engine.regtype.esp) > stackinit||p4.getReg(p4engine.regtype.esp) < stackinit - 5000) {//make this more rigorous
                Label lerr = new Label();
                Label lmsg = new Label();
                lerr.Text = "Error";
                lmsg.Text = "Stack Pointer Out of range";
                lmsg.Width = 70;
                lmsg.Height = 50;
                this.tableLayoutPanel2.RowCount = 1;
                this.tableLayoutPanel2.ColumnCount = 2;
                this.tableLayoutPanel2.Controls.Clear();
                this.tableLayoutPanel2.Controls.Add(lerr, 0, 0);
                this.tableLayoutPanel2.Controls.Add(lmsg, 1, 0);
                return;
            }
			//If Stack Empty..
			//As above - sync with ram.
            if(p4.getReg(p4engine.regtype.esp) == stackinit) {
                Label lerr = new Label();
                Label lmsg = new Label();
                if(hexStackToolStripMenuItem.Checked)lerr.Text = stackinit.ToString("x");
                else lerr.Text = stackinit.ToString();
                lmsg.Text = "Stack is Empty";
                lmsg.Width = 70;
                lmsg.Height = 50;
                this.tableLayoutPanel2.RowCount = 1;
                this.tableLayoutPanel2.ColumnCount = 2;
                this.tableLayoutPanel2.Controls.Clear();
                this.tableLayoutPanel2.Controls.Add(lerr, 0, 0);
                this.tableLayoutPanel2.Controls.Add(lmsg, 1, 0);
                return;
            }
			//If Stack non-empty valid.
			//Get rid of all previous rows and labels
            this.tableLayoutPanel2.Controls.Clear();//optimize this portion...
            this.tableLayoutPanel2.SuspendLayout();
            this.tableLayoutPanel2.ColumnCount = 2;
            int rows = (stackinit - p4.getReg(p4engine.regtype.esp)) / 4;
            this.tableLayoutPanel2.RowCount = rows;
            this.tableLayoutPanel2.Height = rows * 20;
            for(int i = 1; i < rows+1; i++) {
                Label ladd = new Label();
                if(hexStackToolStripMenuItem.Checked) ladd.Text = ((int)(stackinit-i*4)).ToString("x");
                else ladd.Text = ""+(stackinit - i*4);
                ladd.Height = 18;
                ladd.Dock = DockStyle.Fill;
                ladd.TextAlign = ContentAlignment.MiddleLeft;
                this.tableLayoutPanel2.Controls.Add(ladd, 0, i-1);
                Label lval = new Label();
                lval.Width = 70;
                lval.Height = 18;
                lval.Dock = DockStyle.Left ;
                lval.TextAlign = ContentAlignment.MiddleLeft;
                if(hexStackToolStripMenuItem.Checked) lval.Text = ((int)(mr.getIntValue(stackinit - i * 4))).ToString("x");
                else lval.Text = "" + (mr.getIntValue(stackinit - i * 4));
                this.tableLayoutPanel2.Controls.Add(lval, 1, i-1);
            }
            this.tableLayoutPanel2.ResumeLayout();
        }

        private void stepLineToolStripMenuItem_Click(object sender, EventArgs e) {
            button2_Click(btnstepOne, null);
        }
        private string getUsefulCommand(string line) {
            //known limitation: cannot use # inside strings.
            if(line.IndexOf('#') != -1) line = line.Substring(0, line.IndexOf('#'));
            return line.Trim();
        }

		//Toggles the stack viewer panel.
        private void button2_Click_1(object sender, EventArgs e) {
            //stackPanelHolder.Visible = (!stackPanelHolder.Visible);
            if(stackPanelHolder.Visible) {
                stackPanelHolder.Visible = false;
                stackPanel.Height = 30;
            }
            else if(!stackPanelHolder.Visible) {
                stackPanelHolder.Visible = true;
                stackPanel.Height = 100;
                refreshStackList();
            }
            if(button2.ImageIndex == 4) button2.ImageIndex = 5;
            else button2.ImageIndex = 4;
        }
/*            if(e.Button == MouseButtons.Right) {
                contextMenuStrip1.Show(e.X+this.Location.X+richTextBox1.Location.X,
                    e.Y+this.Location.Y+richTextBox1.Location.Y);
            }
 */

        private void menuPaste_Click(object sender, EventArgs e) {
            IDataObject pasteData = Clipboard.GetDataObject();
            //check whether text is available
            if(pasteData.GetDataPresent(DataFormats.Text)) {
                richTextBox1.SelectedText = "";
                richTextBox1.SelectedText = 
                    (string)pasteData.GetData(DataFormats.Text);
            }
            else {
                statusBarItem1.Text = "No Data to paste";
            }
        }

        private void menuCopy_Click(object sender, EventArgs e) {
            //check whether text is available
            if(richTextBox1.SelectedText.Length>0){
                Clipboard.SetDataObject(richTextBox1.SelectedText);
            }
            else {
                statusBarItem1.Text = "No Data to copy";
            }
        }

        private void menuCut_Click(object sender, EventArgs e) {
            //check whether text is available
            if(richTextBox1.SelectedText.Length > 0) {
                Clipboard.SetDataObject(richTextBox1.SelectedText);
                richTextBox1.SelectedText = "";
            }
            else {
                statusBarItem1.Text = "No Data to cut";
            }
        }

        private void menuDelete_Click(object sender, EventArgs e) {
            //check whether text is available
            if(richTextBox1.SelectedText.Length > 0) {
                richTextBox1.SelectedText = "";
            }
        }

        private void menuSelectAll_Click(object sender, EventArgs e) {
            richTextBox1.SelectAll();
        }

        private void onRightPanelPaint(object sender, PaintEventArgs e) {
			int line = p4.getReg(p4engine.regtype.epc);
			if(line < 0 || line > richTextBox1.Lines.Length) return;
            int start = richTextBox1.GetFirstCharIndexFromLine(line);
            currentLocation = richTextBox1.GetPositionFromCharIndex(start);
            //e.Graphics.FillRectangle(Brushes.Yellow, currentLocation.X+5, currentLocation.Y+5, 10, 10);
            e.Graphics.DrawImageUnscaled(imageList1.Images[6], currentLocation.X + 20, currentLocation.Y - 2);
        }

        private void onRichTextScroll(object sender, ScrollEventArgs e) {
            panel1.Refresh();
        }

        private void richtextVscroll(object sender, EventArgs e) {
            panel1.Refresh();
        }

        private void mnuHexRegisters_Click(object sender, EventArgs e) {
            refreshRegisterList();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e) {
            DialogResult dr = openFileDialog1.ShowDialog(this);
            if(dr.Equals(DialogResult.OK)){
                try {
                    filename = openFileDialog1.FileName;
                    filesavedonce = true;
                    StreamReader sr = new StreamReader(openFileDialog1.FileName);
                    StringBuilder sb = new StringBuilder();
                    int i = 0;
                    while((i = sr.Read()) != -1) {
                        sb.Append((char)i);
                    }
                    richTextBox1.Text = sb.ToString();
                    sr.Close();
                }
                catch {
                    MessageBox.Show("Error Reading File");
                    statusBarItem1.Text = "Error Reading File";
                    statusBarItem1.ImageIndex = 2;
                }
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e) {
            if(filesavedonce) {
                saveTo(filename);
            }
            else saveAsToolStripMenuItem_Click(null, null);
        }

        private void saveTo(string _filename) {
            int i = 0;
            string s = richTextBox1.Text;
            try {
                StreamWriter sw = new StreamWriter(_filename);
                for(i = 0; i < s.Length; i++)
                    sw.Write(s[i]);
                sw.Flush();
                sw.Close();
            }
            catch {
                MessageBox.Show("Error Saving File");
                statusBarItem1.Text = "Error Saving File";
            }

        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e) {
            DialogResult dr = saveFileDialog1.ShowDialog(this);
            if(dr.Equals(DialogResult.OK)){
                this.filename = saveFileDialog1.FileName;
                this.filesavedonce = true;
                saveTo(saveFileDialog1.FileName);
            }

        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e) {
            Application.Exit();
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e) {
            DialogResult dr = MessageBox.Show(this, "Any Unsaved Changes Will be Lost! Continue?", "Warning", MessageBoxButtons.YesNoCancel);
            if(dr.Equals(DialogResult.Yes)) {
                richTextBox1.Clear();
                filename = "";
                filesavedonce = false;
            }
        }

        private void undoToolStripMenuItem_Click(object sender, EventArgs e) {
            if(richTextBox1.CanUndo) {
                richTextBox1.Undo();
            }
        }

        private void redoToolStripMenuItem_Click(object sender, EventArgs e) {
            if(richTextBox1.CanRedo) {
                richTextBox1.Redo();
            }
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e) {
            //customControl11.Size = new Size(478, 335);
            //customControl11.Location = new Point(59, 31);
            //customControl11.Visible = true;
        }

        private void button3_Click(object sender, EventArgs e) {
            refreshStackList();
        }

        private void hexStackToolStripMenuItem_Click(object sender, EventArgs e) {
            refreshStackList();
        }

        private void resetToolStripMenuItem_Click(object sender, EventArgs e) {
            p4.setReg(p4engine.regtype.eax, 0);
            p4.setReg(p4engine.regtype.ebp, 0);
            p4.setReg(p4engine.regtype.ebx, 0);
            p4.setReg(p4engine.regtype.ecx, 0);
            p4.setReg(p4engine.regtype.edi, 0);
            p4.setReg(p4engine.regtype.edx, 0);
            p4.setReg(p4engine.regtype.epc, 0);
            p4.setReg(p4engine.regtype.esi, 0);
            p4.setReg(p4engine.regtype.esp, 0);
            labels.Clear();
            mr.Clear();
            refreshRegisterList();
            refreshStackList();
        }

        private void btnCollapse_Click(object sender, EventArgs e) {
            tableLayoutPanel1.RowStyles[1].Height = 10;
        }

        int selectedtabpage = 0;
        private void tabClicked(object sender, EventArgs e) {
            if(splitContainer1.Panel2.Height < 26) {
                splitContainer1.SplitterDistance = splitContainer1.Height - 200;
                selectedtabpage = tabControl1.SelectedIndex;
                return;
            }
            if(tabControl1.SelectedIndex != selectedtabpage) {
                selectedtabpage = tabControl1.SelectedIndex;
                return;
            }
            splitContainer1.SplitterDistance = splitContainer1.Height - 25;
        }

		/// <summary>
		/// Shows or hides the bottom panel having the different messages
		/// </summary>
		/// <param name="tabPage">The panel to show - Compiler Erros(0), Quick Console(1)</param>
		/// <param name="show">Set to true to show the pane, to false to hide it.</param>
		private void showOrHideBottomPane(int tabPage, bool show) {
			tabControl1.SelectedIndex = tabPage;
			selectedtabpage = tabPage;
			if(show) {
				//ensure visibility
				if(splitContainer1.SplitterDistance > splitContainer1.Height - 60) {
					//open it wide
					splitContainer1.SplitterDistance = splitContainer1.Height - 200;
				}
			}
			else {
				//hide it!
				splitContainer1.SplitterDistance = splitContainer1.Height - 25;
			}
		}

		void minimiseTabRequest() {
			showOrHideBottomPane(selectedtabpage, false);
			richTextBox1.Focus();
		}

		void moveCursorToErrorLine(int line) {
			line--;
			if(line < 0 || line > richTextBox1.Lines.Length) return;
			richTextBox1.Select(richTextBox1.GetFirstCharIndexFromLine(line),
				richTextBox1.Lines[line].Length);
			richTextBox1.SelectionBackColor = Color.Pink;
			richTextBox1.DeselectAll();
			richTextBox1.SelectionStart = richTextBox1.GetFirstCharIndexFromLine(line);
			//richTextBox1.SelectionLength = ;
			richTextBox1.Focus();
			richTextBox1.ScrollToCaret();
		}


        private void MyFormLoad(object sender, EventArgs e) {
            richTextBox1.Focus();
        }

		private void contentsToolStripMenuItem_Click(object sender, EventArgs e) {
			//quickConsoleControl1.DisplayPrintf("Hello\\nWorld\\tHere.");
		}


		internal void requestPrint(string t) {
			showOrHideBottomPane(1, true);
			quickConsoleControl1.DisplayPrintf(t);
		}
	}
}