using System;
using System.Collections.Generic;
using System.Text;

namespace Pentium4emulator
{
	enum ErrorTypes
	{
		Error,
		Warning,
		Message
	};
    class MyParser
    {
        Form1 parent;
        public MyParser(Form1 _parent) {
            parent = _parent;
        }
		public delegate void errorOccured(int errormessage, int lineNumber, string param, ErrorTypes type);
		public event errorOccured compileError;

		//returns true on success..
        public bool getGlobals() {
            string line;
            bool datasegstarted = false;
			bool anyerrors = false;
            for(int i = 0; i < parent.programbox.Lines.Length; i++) {
                line = parent.programbox.Lines[i];
                line = getUsefulCommand(line);
                if(datasegstarted) {
                    //if end of data segment...
                    if(line.IndexOf(".text") != -1) break;
                    if(line.Length < 1) continue;
                    //else, try to get the variable name, type and value.
					string[] s = null;
					try {
						s = splitdatadeclaration(line);
					}
					catch {
						compileError(7, (i + 1),"", ErrorTypes.Error);
						anyerrors = true;
						continue;
					}
					if(!s[0].EndsWith(":")) {
						//parent.handleerror("Colon (':') expected, line :" + (i + 1));
						compileError(1, (i + 1),"", ErrorTypes.Error);
						anyerrors = true;
					}
					//remove the colon at the end of s[0].
					else s[0] = s[0].Substring(0, s[0].Length - 1).Trim();
                    //We'll support .string, .space, and .long for now
                    if(".string".Equals(s[1])){
						string ss = s[2].Substring(1, s[2].Length-2).Trim();
                        parent.mr.addString(s[0], ss);
                    }
                    else if(".long".Equals(s[1])) {
                        try {
                            parent.mr.addInt(s[0], getGoodInt(s[2]));
                        }
                        catch { 
							//parent.handleerror("expected an integer on line :" + (i + 1)); 
							compileError(2, (i + 1),"", ErrorTypes.Error);
							anyerrors = true;
						}
                    }
					else if(".space".Equals(s[1])) {
						try {
							parent.mr.malloc(s[0], getGoodInt(s[2]));
						}
						catch {
							compileError(3, (i + 1), "", ErrorTypes.Error);
							anyerrors = true;
						}
					}
					else {
						compileError(4, (i + 1), s[1], ErrorTypes.Error);
						anyerrors = true;
						//parent.handleerror("Unknown datatype: " + s[1]);
					}
				}
				else if(line.IndexOf(".data") != -1) {
					datasegstarted = true;
                }
            }
			return (!anyerrors);
        }
        //A function that can handle normal as well as hex numbers 
        //depending upon whether it has a leading
        //0x or not.
        int getGoodInt(string numberstring) {
            numberstring = numberstring.Trim();
            if(numberstring.StartsWith("0x")) {
                numberstring = numberstring.Substring(2);
                return int.Parse(numberstring, System.Globalization.NumberStyles.HexNumber);
            }
            if(numberstring.StartsWith("-0x")) {
                numberstring = numberstring.Substring(3);
                return -(int.Parse(numberstring, System.Globalization.NumberStyles.HexNumber));
            }
            else return int.Parse(numberstring);
        }
        public void findLabels() {
            string line;
            bool codesegstarted = false;
            for(int i = 0; i < parent.programbox.Lines.Length; i++) {
                line = parent.programbox.Lines[i];
                line = getUsefulCommand(line);
                if(codesegstarted) {
                    //try and find labels
                    if(line.EndsWith(":")) {
                        parent.labels.Add(line.Substring(0, line.Length - 1), i + 1);
                    }
                }
                if(line.IndexOf(".text") != -1) codesegstarted = true;
            }
        }
        private string getUsefulCommand(string line) {
            //known limitation: cannot use # inside strings.
            if(line.IndexOf('#') != -1) line = line.Substring(0, line.IndexOf('#'));
            return line.Trim();
        }

        private string[] splitdatadeclaration(string s) {
            //declare a string[] = { name, type, value}
            string[] reply = new string[3];
            char ch;
            int i = 0;
            s = s.Trim();
            if(s.Equals("")) return null;
            while(!Char.IsWhiteSpace(ch = s[i++])) reply[0] += ch;
            s = s.Substring(i, s.Length - i);
            s = s.Trim();
            i = 0;
            while(!Char.IsWhiteSpace(ch = s[i++])) reply[1] += ch;
            s = s.Substring(i, s.Length - i);
            s = s.Trim();
            reply[2] = s;
            return reply;
        }

		
		/// <summary>
		/// A function that parses and executes one line of assembler code.
		/// </summary>
		/// <param name="command">The command (stripped of comments) that is to be executed</param>
		/// <param name="lineNumber">The line number on which this command exists
		/// Useful for locating the errors.</param>
		/// <returns>An array of strings {opcode, param1, param2}</returns>
		public void parseAndExecuteLine(string command, int lineNumber){
			command = command.Trim();
			string opcode = "", operand1 = "", operand2 = "";
			int i=0;
			char ch=' ';
			bool inbracket = false;
			
			//Get the command string (movl, pushl, etc.)
			while(i < command.Length) {
				ch = command[i++];
				if(char.IsWhiteSpace(ch)) break;
				opcode += ch;
			}

			//Get the first operand
			while(i < command.Length) {
				ch = command[i++];
				if(ch == ',' && !(inbracket)) break;
				if(ch == '(') {
					if(inbracket) throw new Exception("14");
					inbracket = true;
				}
				if(ch == ')') {
					if(!inbracket) throw new Exception("15");
					inbracket = false;
				}
				operand1 += ch;
			}
			operand1 = operand1.Trim();

			//Get the second operand
			while(i < command.Length) {
				ch = command[i++];
				if(ch == ',' && !(inbracket)) break;
				if(ch == '(') {
					if(inbracket) throw new Exception("14");
					inbracket = true;
				}
				if(ch == ')') {
					if(!inbracket) throw new Exception("15");
					inbracket = false;
				}
				operand2 += ch;
			}
			operand2 = operand2.Trim();
			Operand op1 = null, op2 = null;
			try{
				op1 = new Operand(operand1);
				op2 = new Operand(operand2);
			}catch(Exception ex){
				try {
					string[] sarr2 = ex.Message.Split(new char[] { ',' });
					int ii = int.Parse(sarr2[0]);
					if(sarr2.Length > 1) {
						compileError(ii, lineNumber, sarr2[1], ErrorTypes.Error);
					}
					else {
						compileError(ii, lineNumber, "", ErrorTypes.Error);
					}
				}
				catch {
					compileError(0, lineNumber, ex.Message, ErrorTypes.Error);
				}
				return;
			}

			try {
				this.parent.p4.executeOpcode(opcode, op1, op2);
			}
			catch(Exception exx) {
				try {
					string[] sarr2 = exx.Message.Split(new char[] { ',' });
					int ii = int.Parse(sarr2[0]);
					if(sarr2.Length > 1) {
						compileError(ii, lineNumber, sarr2[1], ErrorTypes.Error);
					}
					else {
						compileError(ii, lineNumber, "", ErrorTypes.Error);
					}
				}
				catch {
					compileError(0, lineNumber, exx.Message, ErrorTypes.Error);
				}
			}
			
		}
 
    }
}
