using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace Pentium4emulator
{
	public partial class QuickConsoleControl : UserControl
	{

		public event MyTableViewer.myEventHandler minimiseRequested;

		public QuickConsoleControl() {
			InitializeComponent();
		}
		/// <summary>
		/// Clears all the text in the control
		/// </summary>
		public void Clear() {
			this.textBox1.Clear();
			this.textBox2.Clear();
		}

		/// <summary>
		/// The Event handler for the event when the user clicks the minimise button
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void minimiseButton_Click(object sender, EventArgs e) {
			minimiseRequested();
		}

		/// <summary>
		/// The Event handler called when the control gets focus
		/// This sets the 'titlebar' to blue color
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void myControlGotFocus(object sender, EventArgs e) {
			this.panel1.BackColor = Color.FromArgb(53,116,215);
			this.label1.ForeColor = Color.White;
			this.minimiseButton.ForeColor = Color.White;
		}

		/// <summary>
		/// The function called when the control loses focus.
		/// This sets the 'titlebar' to light brown
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void myControlLostFocus(object sender, EventArgs e) {
			this.panel1.BackColor = SystemColors.InactiveBorder;
			this.label1.ForeColor = SystemColors.ControlText;
			this.minimiseButton.ForeColor = SystemColors.ControlText;
		}

		private void InputKeyDown(object sender, KeyEventArgs e) {
			if(e.KeyCode == Keys.Enter) {
				//
			}
		}

		/// <summary>
		/// A function that 'displays' on the screen the string passed to it
		/// after replacing escaped sequences.
		/// </summary>
		/// <param name="s">The string to display. May contain escape sequences</param>
		public void DisplayPrintf(string s) {
			StringBuilder sb = new StringBuilder(s.Length);

			#region Build String replacing escaped sequences
			//Whether a backslash has been encountered.
			bool inescape = false;
			for(int i = 0; i < s.Length; i++) {
				if(inescape) {
					//This is a special character.
					switch(s[i]) {
						case '\\':
							sb.Append('\\');
							break;
						case 'n':
							sb.Append("\r\n");//Line break
							break;
						case 't':
							sb.Append('\t');
							break;
						case 'r':
							sb.Append("\r\n");//Think of somethings here..
							break;
						case '"':
							sb.Append('"');
							break;
						default:
							throw new Exception("17, \\" + s[i]);
					}
					inescape = false;
					continue;
				}
				//If not an escaped character,
				if(s[i] == '\\') {
					inescape = true;
					continue;
				}
				else {
					sb.Append(s[i]);
				}

			} 
			#endregion

			textBox1.AppendText(sb.ToString());

		}

	}
}
