using System;
using System.Collections.Generic;
using System.Text;

namespace Pentium4emulator
{
    /// <summary>
    /// This is a ram.
    /// This has a seperate integer segment, string segment and stack segments.
    /// </summary>
    public class myRam
    {
        int[] stkSeg = new int[1024];//Memory locations 10000 - 11023
        int[] iSeg = new int[1024];//Memory locations 20000 - 21023
        string[] szSeg = new string[128];//Memory locations 30000 - 30127

        const int stkOffset = 10000;
        const int iOffset = 20000;
        const int szOffset = 30000;

        //The 'Variable Allocation Table'
        private Dictionary<string, int> vat = new Dictionary<string, int>();

        //The number of elements already entered into the memory
        private int iLength = 0;
        private int szLength = 0;

        //Add a string element to the Ram
        public int addString(string name, string value) {
            szSeg[szLength] = value;
            int reply = szOffset + szLength;
            vat.Add(name, reply);
            szLength++;
            return (reply*4);//return ptr to string.
        }

        //Add an integer element to the Ram
        public int addInt(string name, int value) {
            iSeg[iLength] = value;
            int reply = iOffset + iLength;
            vat.Add(name, reply);
            iLength++;
            return (reply*4);
        }

        //Allocate a region of the memory
        public int malloc(string name, int length) {
            int reply = iOffset + iLength;
            vat.Add(name, reply);
            iLength += length;
            return reply*4;
        }

        //Save an int to a named variable
        public int saveValue(string name, int value) {
            int destination = vat[name];
            if(destination >= stkOffset && destination < iOffset) stkSeg[destination - stkOffset] = value;
            else if(destination >= iOffset && destination < szOffset) iSeg[destination - iOffset] = value;
            else {
                //error
            }
            return value;
        }

        //Save a string to a named variable
        public string saveValue(string name, string value) {
            int destination = vat[name];
            if(destination >= szOffset) szSeg[destination - szOffset] = value;
            else {
                //error
            }
            return value;
        }

        //Save an int to a pointer 
        public int saveValue(int ptr, int value) {
            if(ptr % 4 != 0) {
                //error - not yet implemented.
                return 0;
            }
            int destination = ptr/4;
            if(destination >= stkOffset && destination < iOffset) stkSeg[destination - stkOffset] = value;
            else if(destination >= iOffset && destination < szOffset) iSeg[destination - iOffset] = value;
            else {
                //error
            }
            return value;
        }

        //Save a string to a pointer
        public string saveValue(int ptr, string value) {
            if(ptr % 4 != 0) {
                //error - not yet implemented.
                return null;
            }
            int destination = ptr/4;
            if(destination >= szOffset) szSeg[destination - szOffset] = value;
            else {
                //error
            }
            return value;
        }

        //Get an int from a named variable
        public int getIntValue(string name) {
            int destination = vat[name];
            if(destination >= stkOffset && destination < iOffset)
                return stkSeg[destination - stkOffset];
            else if(destination >= iOffset && destination < szOffset)
                return iSeg[destination - iOffset];
            else {
                //error
            }
            return 0;
        }

        //Get an int from a pointer
        public int getIntValue(int ptr) {
            if(ptr % 4 != 0) {
                //error - not yet implemented.
                return 0;
            }
            int destination = ptr/4;
            if(destination >= stkOffset && destination < iOffset)
                return stkSeg[destination - stkOffset];
            else if(destination >= iOffset && destination < szOffset)
                return iSeg[destination - iOffset];
            else {
                //error
            }
            return 0;
        }

        //Get a string from a named variable
        public string getStringValue(string name) {
            int destination = vat[name];
            if(destination >= szOffset)
                return szSeg[destination - szOffset];
            else {
                //error
                return null;
            }
        }

        //Get a string from a pointer
        public string getStringValue(int ptr) {
			if(ptr % 4 != 0) {
				//error - not yet implemented.
				return null;
			}
			int destination = ptr / 4;
			//int destination = ptr;
            if(destination >= szOffset)
                return szSeg[destination - szOffset];
            else {
                //error
                return null;
            }
        }

        //Get the pointer from a named variable
        public int getPointer(string name) {
            return vat[name]*4;
        }
        

        internal void Clear() {
             vat.Clear();
        }
    }
}
