using System;
using System.Collections.Generic;
using System.Text;

namespace Pentium4emulator
{
    public class p4engine
    {
        public p4engine(myRam _ram, Form1 _parent) {
            this.ram = _ram;
            this.parent = _parent;
            registers.Add(regtype.eax, 0);
            registers.Add(regtype.ebx, 0);
            registers.Add(regtype.ecx, 0);
            registers.Add(regtype.edx, 0);
            registers.Add(regtype.esp, 0);
            registers.Add(regtype.ebp, 0);
            registers.Add(regtype.esi, 0);
            registers.Add(regtype.edi, 0);
            registers.Add(regtype.epc, 0);
            flag.Add(flags.auxiliarycarry, false);
            flag.Add(flags.carry, false);
            flag.Add(flags.df, false);
            flag.Add(flags.iflag, false);
            flag.Add(flags.nt, false);
            flag.Add(flags.overflow, false);
            flag.Add(flags.parity, false);
            flag.Add(flags.rf, false);
            flag.Add(flags.sign, false);
            flag.Add(flags.tf, false);
            flag.Add(flags.zero, false);
        }


        const int codesegment = 60000;
        Form1 parent;
        /// <summary>
        /// The ram for the system.
        /// </summary>
        myRam ram;
        /// <summary>
        /// The list of registers
        /// </summary>
        public enum regtype
        {
            eax,
            ebx,
            ecx,
            edx,
            esi,
            edi,
            esp,
            ebp,
            epc
        };
        /// <summary>
        /// The registers of this machine.
        /// </summary>
        Dictionary<regtype, int> registers = new Dictionary<regtype, int>(10);
        /// <summary>
        /// The flags of this cpu
        /// </summary>
        enum flags { overflow, sign, zero, auxiliarycarry, parity, carry, tf, iflag, df, nt, rf };
        Dictionary<flags, bool> flag = new Dictionary<flags, bool>(12);

        public int getReg(regtype r) {
            return registers[r];
        }


        private void resetflags() {
            flag[flags.auxiliarycarry] = false;
            flag[flags.carry] = false;
            flag[flags.df] = false;
            flag[flags.iflag] = false;
            flag[flags.nt] = false;
            flag[flags.overflow] = false;
            flag[flags.parity] = false;
            flag[flags.rf] = false;
            flag[flags.sign] = false;
            flag[flags.tf] = false;
            flag[flags.zero] = false;           
        }
        public void setReg(regtype r, int i) {
            registers[r] = i;
        }

        int getGoodInt(string numberstring) {
            numberstring = numberstring.Trim();
            if(numberstring.StartsWith("0x")) {
                numberstring = numberstring.Substring(2);
                return int.Parse(numberstring, System.Globalization.NumberStyles.HexNumber);
            }
            if(numberstring.StartsWith("-0x")) {
                numberstring = numberstring.Substring(3);
                return -(int.Parse(numberstring, System.Globalization.NumberStyles.HexNumber));
            }
            else return int.Parse(numberstring);
        }



		/// <summary>
		/// The function that performs the same way a P4 would do
		/// </summary>
		/// <param name="opcode">The 'command' like 'movl', 'pushl', etc. no extra qhitespace allowed</param>
		/// <param name="op1">The first operand</param>
		/// <param name="op2">The second operand</param>
		internal void executeOpcode(string opcode, Operand op1, Operand op2) {
			int result = 0;
			switch(opcode) {
				case "movl":
					result = getValueFrom(op1);
					saveValueTo(op2, result);
					break;
				case "addl":
					result = getValueFrom(op1);
					result += getValueFrom(op2);
					saveValueTo(op2,result);
					break;
				case "subl":
					result = getValueFrom(op2);
					result -= getValueFrom(op1);
					saveValueTo(op2, result);
					break;
				case "imul":
					result = getValueFrom(op1);
					result *= getReg(regtype.eax);
					setReg(regtype.eax, result);
					break;
				case "andl":
					result = getValueFrom(op1);
					result &= getValueFrom(op2);
					saveValueTo(op2, result);
					break;
				case "orl":
					result = getValueFrom(op1);
					result |= getValueFrom(op2);
					saveValueTo(op2, result);
					break;
				case "xorl":
					result = getValueFrom(op1);
					result ^= getValueFrom(op2);
					saveValueTo(op2, result);
					break;
				case "negl":
					result = getValueFrom(op1);
					result = -result;
					saveValueTo(op1, result);
					break;
				case "incl":
					result = getValueFrom(op1);
					result++;
					saveValueTo(op1, result);
					break;
				case "decl":
					result = getValueFrom(op1);
					result--;
					saveValueTo(op1, result);
					break;
				case "shll":
					result = getValueFrom(op2);
					result = result<<(getValueFrom(op1));
					saveValueTo(op2, result);
					break;
				case "sall":
					result = getValueFrom(op2);
					result = result << (getValueFrom(op1));
					saveValueTo(op2, result);
					break;
				case "shrl":
					result = getValueFrom(op2);
					result = result >> (getValueFrom(op1));
					saveValueTo(op2, result);
					break;
				case "sarl":
					if(getValueFrom(op1) == 0) break;
					result = getValueFrom(op2);
					if(result < 0) {
						result = result >> 1;
						result &= 0x7fffffff;
						result = result >> (getValueFrom(op1) - 1);
					}
					else 
						result = result >> (getValueFrom(op1));
					saveValueTo(op2, result);
					break;
				case "cmpl":
					result = getValueFrom(op2);
					result -= getValueFrom(op1);
					if(result == 0)
						flag[flags.zero] = true;
					else 
						flag[flags.zero] = false;
					if(result < 0)
						flag[flags.carry] = true;
					else
						flag[flags.carry] = false;
					break;
				case "pushl":
					result = getValueFrom(op1);
					registers[regtype.esp] -= 4;
					parent.mr.saveValue(getReg(regtype.esp), result);
					break;
				case "popl":
					result = parent.mr.getIntValue(getReg(regtype.esp));
					registers[regtype.esp] += 4;
					saveValueTo(op1, result);
					break;
				case "call":
					if(op1.AddressingMode != addressingMode.namedVariable)
						throw new Exception("18");
					if(op1.name.Equals("_printf")) {
						handle_printf();
						break;
					}
					if(op1.name.Equals("_scanf")) {
						handle_scanf();
						break;
					}
					registers[regtype.esp] -= 4;
					parent.mr.saveValue(getReg(regtype.esp), codesegment + getReg(regtype.epc));
					try { result = parent.labels[op1.name]; }
					catch { throw new Exception("10"); }
					registers[regtype.epc] = result - 1;//epc will be incremented at the end
					break;
				case "ret":
					result = parent.mr.getIntValue(registers[regtype.esp]);
					registers[regtype.esp] += 4;
					if(result < codesegment) throw new Exception("16");
					registers[regtype.epc] = result - codesegment;
					break;
				case "jmp":
					try { result = parent.labels[op1.name]; }
					catch { throw new Exception("10"); }
					registers[regtype.epc] = result - 1;//epc will be incremented at the end
					break;
				case "jge":
					if(!flag[flags.carry]) {
						try { result = parent.labels[op1.name]; }
						catch { throw new Exception("10"); }
						registers[regtype.epc] = result - 1;//epc will be incremented at the end
					}
					break;
				case "jg":
					if(!flag[flags.carry] && !flag[flags.zero]) {
						try { result = parent.labels[op1.name]; }
						catch { throw new Exception("10"); }
						registers[regtype.epc] = result - 1;//epc will be incremented at the end
					}
					break;
				case "je":
					if(flag[flags.zero]) {
						try { result = parent.labels[op1.name]; }
						catch { throw new Exception("10"); }
						registers[regtype.epc] = result - 1;//epc will be incremented at the end
					}
					break;
				case "jle":
					if(flag[flags.carry]||flag[flags.zero]) {
						try { result = parent.labels[op1.name]; }
						catch { throw new Exception("10"); }
						registers[regtype.epc] = result - 1;//epc will be incremented at the end
					}
					break;
				case "jl":
					if(flag[flags.carry] && !flag[flags.zero]) {
						try { result = parent.labels[op1.name]; }
						catch { throw new Exception("10"); }
						registers[regtype.epc] = result - 1;//epc will be incremented at the end
					}
					break;
				case "jne":
					if(!flag[flags.zero]) {
						try { result = parent.labels[op1.name]; }
						catch { throw new Exception("10"); }
						registers[regtype.epc] = result - 1;//epc will be incremented at the end
					}
					break;
				case "jz":
					if(flag[flags.carry] && !flag[flags.zero]) {
						try { result = parent.labels[op1.name]; }
						catch { throw new Exception("10"); }
						registers[regtype.epc] = result - 1;//epc will be incremented at the end
					}
					break;
				case "jnz":
					if(!flag[flags.zero]) {
						try { result = parent.labels[op1.name]; }
						catch { throw new Exception("10"); }
						registers[regtype.epc] = result - 1;//epc will be incremented at the end
					}
					break;
				case "leal":
					result = getPointerTo(op1);
					saveValueTo(op2, result);
					break;
			}
			registers[regtype.epc]++;
		}

		private void handle_scanf() {
			throw new Exception("The method or operation is not implemented.");
		}

		private void handle_printf() {
			//First get the format string.
			//Note that we have broken out of the call op-code default handler, so we dont have
			//  a return address at the top of stack as we would have otherwise.
			string s = parent.mr.getStringValue(
				parent.mr.getIntValue(registers[regtype.esp]));
			//walk through the format string hunting for format characters...
			#region Build String replacing format characters
			//Whether a percentage has been encountered.
			bool inescape = false;
			//The depth to which we have gone inside the stack
			int ij = 4;
			//The String Builder used to construct the final string
			StringBuilder sb = new StringBuilder();
			for(int ii = 0; ii < s.Length; ii++) {
				if(inescape) {
					//This is a special character.
					switch(s[ii]) {
						case 'd':
							sb.Append(parent.mr.getIntValue(registers[regtype.esp] + ij));
							//ii += 4;  <-------------------------- Why did I ever do this ????
							break;
						case '%':
							sb.Append('%');
							break;
						default:
							throw new Exception("17, %" + s[ii]);
					}
					inescape = false;
					continue;
				}
				//If not an escaped character,
				if(s[ii] == '%') {
					inescape = true;
					continue;
				}
				else {
					sb.Append(s[ii]);
				}

			}
			#endregion
			string t = sb.ToString();
			parent.requestPrint(t);


		}

		/// <summary>
		/// A helper function for the leal op-code. Gets the pointer to a Operand.
		/// </summary>
		/// <param name="op1">The Operand whose pointer we want to find</param>
		/// <returns>Pointer to the Operand.</returns>
		private int getPointerTo(Operand op1) {
			switch(op1.AddressingMode) {
				case addressingMode.nullOperand:
					throw new Exception("8");
				case addressingMode.immediateValue:
					throw new Exception("11");
				case addressingMode.immediateRegister:
					throw new Exception("12");
				case addressingMode.indirectByRegister:
					return getReg(op1.reg1);
				case addressingMode.offsetBase:
					return (getReg(op1.reg1) + op1.imm1);
				case addressingMode.baseOffset:
					return (getReg(op1.reg1) + getReg(op1.reg2));
				case addressingMode.baseOffsetScaling:
					return (getReg(op1.reg1) + op1.imm1 * getReg(op1.reg2));
				case addressingMode.offsetBaseOffsetScaling:
					return (getReg(op1.reg1) + op1.imm1 * getReg(op1.reg2) + op1.imm2);
				case addressingMode.namedVariable:
					return (parent.mr.getPointer(op1.name));
			}
			throw new Exception("9");
		}

		/// <summary>
		/// Gets the value of an Operand
		/// </summary>
		/// <param name="op1">The Operand whose value we want to find</param>
		/// <returns>The value of the operand.</returns>
		private int getValueFrom(Operand op1) {
			int ptr = 0, offset = 0;
			switch(op1.AddressingMode) {
				case addressingMode.nullOperand:
					throw new Exception("8");//Error Number 8
				case addressingMode.immediateValue:
					return op1.imm1;
				case addressingMode.immediateRegister:
					return getReg(op1.reg1);
				case addressingMode.indirectByRegister:
					ptr = getReg(op1.reg1);
					return(parent.mr.getIntValue(ptr));
				case addressingMode.offsetBase:
					ptr = getReg(op1.reg1);
					ptr += op1.imm1;
					return (parent.mr.getIntValue(ptr));
				case addressingMode.baseOffset:
					ptr = getReg(op1.reg1);
					offset = getReg(op1.reg2);
					return (parent.mr.getIntValue(ptr + offset));
				case addressingMode.baseOffsetScaling:
					ptr = getReg(op1.reg1);
					offset = getReg(op1.reg2);
					return (parent.mr.getIntValue(ptr + op1.imm1 * offset));
				case addressingMode.offsetBaseOffsetScaling:
					ptr = getReg(op1.reg1);
					offset = getReg(op1.reg2);
					return (parent.mr.getIntValue(ptr + op1.imm1 * offset + op1.imm2));
				case addressingMode.namedVariable:
					return (parent.mr.getIntValue(op1.name));
			}
			throw new Exception("9");//Error number 9
		}

		/// <summary>
		/// Saves the specified value to the Operand specified.
		/// </summary>
		/// <param name="op1">The target Operand where to save the value</param>
		/// <param name="value">The value to save to the Operand</param>
		private void saveValueTo(Operand op1, int value) {
			int ptr = 0, offset = 0;
			switch(op1.AddressingMode) {
				case addressingMode.nullOperand:
					throw new Exception("8");//Error Number 8
				case addressingMode.immediateValue:
					throw new Exception("5");
				case addressingMode.immediateRegister:
					setReg(op1.reg1, value);
					return;
				case addressingMode.indirectByRegister:
					ptr = getReg(op1.reg1);
					parent.mr.saveValue(ptr,value);
					return;
				case addressingMode.offsetBase:
					ptr = getReg(op1.reg1);
					ptr += op1.imm1;
					parent.mr.saveValue(ptr, value);
					return;
				case addressingMode.baseOffset:
					ptr = getReg(op1.reg1);
					offset = getReg(op1.reg2);
					parent.mr.saveValue(ptr + offset,value);
					return;
				case addressingMode.baseOffsetScaling:
					ptr = getReg(op1.reg1);
					offset = getReg(op1.reg2);
					parent.mr.saveValue(ptr + op1.imm1 * offset, value);
					return;
				case addressingMode.offsetBaseOffsetScaling:
					ptr = getReg(op1.reg1);
					offset = getReg(op1.reg2);
					parent.mr.saveValue(ptr + op1.imm1 * offset + op1.imm2, value);
					return;
				case addressingMode.namedVariable:
					parent.mr.saveValue(op1.name,value);
					return;
			}
			throw new Exception("9");//Error number 9
		}
	}


}
